/*********************************************************************
 * main.c - Main firmware (ATmega328p version)                          *
 * $Id: $
 * Version 1.98                                                     *
 *********************************************************************
 * electron-keyboard is Copyright(C) 2015 Kernel Crash
 * It is heavily based on c64key (Copyright (C) 2006-2007 Mikkel 
 * Holm Olsen which is based on based on HID-Test by 
 * Christian Starkjohann, Objective Development *
 *********************************************************************
 * Kernel Crash's Acorn Electron 64 USB Keyboard 
 * is free software; you can redistribute it and/or modify it under  *
 * the terms of the OBDEV license, as found in the licence.txt file. *
 *                                                                   *
 * electron-keyboard is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
 * OBDEV license for further details.                                *
 *********************************************************************/

#define TCCR0   TCCR0B

#include "key_electron_uk_uk.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <string.h>


#include "usbdrv.h"
#define DEBUG_LEVEL 0
#include "oddebug.h"

/* Hardware documentation:

 * My atmega328p acorn electron implementation
 * ROW OUTPUTS (need 14 pins)
 * PB0 to PB5, PC4, PC5, PD1, PD3, PD4, PD5, PD6, PD7
 * PB0 is the esc, caplock, ctrl, shift row, PB1 is 1,q,a,z and so forth
 *
 * PC0 to PC3 as column inputs. PC0 is the rightmost column, PC3 is leftmost
 * as per the Acorn Electron Service manual diagram of the key matrix
 *
 * PD0 is D-, PD2 is D+
 * 16MHz crystal on XTAL1/XTAL2
 * reset on PC6. With 10K pullup to Vcc
 *
 * USB Connector:
 * -------------
 *  1 (red)    +5V
 *  2 (white)  DATA-
 *  3 (green)  DATA+
 *  4 (black)  GND
 *    
 *
 *                                                  ATMega328p
 *      USB        GND        |                     
 *                                     
 *               |\ |    |\ |
 *      (5V)-----| >|----| >|-----------+---------  VCC
 *               |/ |    |/ |           |
 *              1N4148's      +--[1k5]--+
 *                            |                     
 *                            |
 *      (D-)-------+----------+--------[68r]------- PD0
 *                  
 *      (D+)-------------+-------------[68r]------- PD2/INT0
 *                          
 *
 *
 */

/* The LED states */
#define LED_NUM     0x01
#define LED_CAPS    0x02
#define LED_SCROLL  0x04
#define LED_COMPOSE 0x08
#define LED_KANA    0x10


/* Originally used as a mask for the modifier bits, but now also
   used for other x -> 2^x conversions (lookup table). */
const char modmask[8] PROGMEM = {
    0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80
  };


/* USB report descriptor (length is defined in usbconfig.h)
   This has been changed to conform to the USB keyboard boot
   protocol */
const PROGMEM char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH]  = {
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,                    // USAGE (Keyboard)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x08,                    //   REPORT_COUNT (8)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x81, 0x03,                    //   INPUT (Cnst,Var,Abs)
    0x95, 0x05,                    //   REPORT_COUNT (5)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x05, 0x08,                    //   USAGE_PAGE (LEDs)
    0x19, 0x01,                    //   USAGE_MINIMUM (Num Lock)
    0x29, 0x05,                    //   USAGE_MAXIMUM (Kana)
    0x91, 0x02,                    //   OUTPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x03,                    //   REPORT_SIZE (3)
    0x91, 0x03,                    //   OUTPUT (Cnst,Var,Abs)
    0x95, 0x06,                    //   REPORT_COUNT (6)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
    0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
    0xc0                           // END_COLLECTION  
};

/* This buffer holds the last values of the scanned keyboard matrix */
static uchar bitbuf[NUMROWS]={0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff};

/* The ReportBuffer contains the USB report sent to the PC */
static uchar reportBuffer[8];    /* buffer for HID reports */
static uchar idleRate;           /* in 4 ms units */
static uchar protocolVer=1;      /* 0 is the boot protocol, 1 is report protocol */

static void hardwareInit(void) {
  PORTB = 0x3F;   /* Port B pullups on bits 0 to 5 */
  /* re DDR registers 0 bit means input, 1 bit means output */
  DDRB  = 0x00;   

  PORTC = 0xff;   /* activate all pull-ups */
  DDRC  = 0x00;   /* all pins input */
  
  PORTD = 0xfa;   /* 1111 1010 bin: activate pull-ups except on USB lines */
  DDRD = 0x02;    /* 0000 0010 bin: remove USB reset condition */

  /* configure timer 0 for a rate of 12M/(1024 * 256) = 45.78 Hz (~22ms) , 61Hz (~16ms) with 16MHz crystal*/
  TCCR0B = 5;      /* timer 0 prescaler: 1024 */
}

/* Table for decoding bit positions of the last three rows */
const char extrows[5] PROGMEM = { 0x08, 0x10, 0x20, 0x40, 0x80 };


/* This function scans the entire keyboard, debounces the keys, and
   if a key change has been found, a new report is generated, and the
   function returns true to signal the transfer of the report. */
static uchar scankeys(void) {
  uchar reportIndex=1; /* First available report entry is 2 */
  uchar retval=0;
  uchar row,data,key, modkeys;
  volatile uchar col, mask;
  static uchar debounce=5;

  for (row=0;row<NUMROWS;++row) { /* Scan all rows */
    PORTB = 0x3F;   /* Port B are row drivers */
    DDRB  = 0x00;   /* B is initially set to all inputs. As we scan, we will change one at a time to output */  
    DDRD&= 0b00000101; /* dont touch PD0 and PD2 */
    PORTD|=0b11111010; /* pull-up enable */
    DDRC&=0b11001111; /* change port C 4 and 5 to inputs */
    PORTC|=0b00110000; /* pull-up enable */
    /* TODO
       if row is 0 to 5 then send LOW one bit on PB0 to 5.
       if row is 6 or 7 then send LOW one bit on PC4 or PC5
       if row is 8, 9, 10, 11, 12, 13, 14 then send LOW either PD1, PD3, PD4, PD5, PD6, PD7
    */
    switch (row) {
       case 0:
       case 1:
       case 2:
       case 3:
       case 4:
       case 5:
          data=pgm_read_byte(&modmask[row]);
          DDRB=data;   // eg. 00010000 for PB4 as output
          PORTB=~data; // eg. 11101111 to set PB4 low
          break;
       case 6:
          DDRC|=0x10;
          PORTC&=~0x10;
          break;
       case 7:
          DDRC|=0x20;
          PORTC&=~0x20;
          break;
       case 8:
          DDRD|=0x02;
          PORTD&=~0x02;
          break;
       case 9:
       case 10:
       case 11:
       case 12:
       case 13:
         data=pgm_read_byte(&extrows[row-9]);
         DDRD|=data;
         PORTD&=~data;
    }
    
    _delay_us(30); /* Used to be small loop, but the compiler optimized it away ;-) */
    
    // read low 4 bits of PORTC
    data=(PINC&0x0F);
    if (data^bitbuf[row]) { 
      debounce=10; /* If a change was detected, activate debounce counter */
    }
    // Hack for break key. Due to the way it is hacked in, the other keys in the same
    // row will register at the same time. We just force them to not register
    if ((row == 11) && ((data&0x08)==0)) {
       data=0x07;
    }
    bitbuf[row]=data; /* Store the result */
  }

  if (debounce==1) { /* Debounce counter expired */
    modkeys=0;
    memset(reportBuffer,0,sizeof(reportBuffer)); /* Clear report buffer */
    for (row=0;row<NUMROWS;++row) { /* Process all rows for key-codes */
      data=bitbuf[row]; /* Restore buffer */
      
      if (data!=0xFF) { /* Anything on this row? - optimization */
        for (col=0,mask=1;col<4;++col,mask<<=1) { /* yes - check individual bits */
          if (!(data&mask)) { /* Key detected */
            key=pgm_read_byte(&keymap[row][col]); /* Read keyboard map */
            if (key>KEY_Special) { /* Special handling of shifted keys */
              /* Modifiers have not been decoded yet - handle manually */
              uchar keynum=key-(KEY_Special+1);
              if ((bitbuf[0]&0b00001000)==0) { /* if bit3 of row0 is low SHIFT */
                key=pgm_read_byte(&spec_keys[keynum][2]); /* Shifted */
                modkeys=pgm_read_byte(&spec_keys[keynum][3]);
              } else if ((bitbuf[0]&0b00000010)==0) { /* if bit1 of row0 is low then capslock/fn */
                key=pgm_read_byte(&spec_keys[keynum][4]); /* Alt */
                modkeys=pgm_read_byte(&spec_keys[keynum][5]);
              } else {
                key=pgm_read_byte(&spec_keys[keynum][0]); /* Unmodified */
                modkeys=pgm_read_byte(&spec_keys[keynum][1]);
              }
            } else if (key>KEY_Modifiers) { /* Is this a modifier key? */
              reportBuffer[0]|=pgm_read_byte(&modmask[key-(KEY_Modifiers+1)]);
              key=0;
            }
            if (key) { /* Normal keycode should be added to report */
              if (++reportIndex>=sizeof(reportBuffer)) { /* Too many keycodes - rollOver */
                if (!(retval&0x02)) { /* Only fill buffer once */
                  memset(reportBuffer+2, KEY_errorRollOver, sizeof(reportBuffer)-2);
                  retval|=2; /* continue decoding to get modifiers */
                }
              } else {
                reportBuffer[reportIndex]=key; /* Set next available entry */
              }
            }
          }
        }
      }
    }
    if (modkeys&0x80) { /* Clear RSHIFT */
      reportBuffer[0]&=~0x20;
    }
    if (modkeys&0x08) { /* Clear L_ALT */
      reportBuffer[0]&=~0x04;
    }
    reportBuffer[0]|=modkeys&0x77; /* Set other modifiers */

    retval|=1; /* Must have been a change at some point, since debounce is done */
  }
  if (debounce) debounce--; /* Count down, but avoid underflow */
  return retval;
}

uchar expectReport=0;
uchar LEDstate=0;

uchar usbFunctionSetup(uchar data[8]) {
  usbRequest_t *rq = (void *)data;
  usbMsgPtr = reportBuffer;
  if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
    if(rq->bRequest == USBRQ_HID_GET_REPORT){  
      /* wValue: ReportType (highbyte), ReportID (lowbyte) */
      /* we only have one report type, so don't look at wValue */
      return sizeof(reportBuffer);
    }else if(rq->bRequest == USBRQ_HID_SET_REPORT){
      if (rq->wLength.word == 1) { /* We expect one byte reports */
        expectReport=1;
        return 0xFF; /* Call usbFunctionWrite with data */
      }  
    }else if(rq->bRequest == USBRQ_HID_GET_IDLE){
      usbMsgPtr = &idleRate;
      return 1;
    }else if(rq->bRequest == USBRQ_HID_SET_IDLE){
      // idleRate is the number 4ms intervals till we have to do something
      idleRate = rq->wValue.bytes[1];
    }else if(rq->bRequest == USBRQ_HID_GET_PROTOCOL) {
      if (rq->wValue.bytes[1] < 1) {
        protocolVer = rq->wValue.bytes[1];
      }
    }else if(rq->bRequest == USBRQ_HID_SET_PROTOCOL) {
      usbMsgPtr = &protocolVer;
      return 1;
    }
  }
  return 0;
}

uchar usbFunctionWrite(uchar *data, uchar len) {
  if ((expectReport)&&(len==1)) {
    LEDstate=data[0]; /* Get the state of all 5 LEDs */
    if (LEDstate&LED_CAPS) { /* Check state of CAPS lock LED */
      PORTD|=0x02;
    } else {
      PORTD&=~0x02;
    }
    expectReport=0;
    return 1;
  }
  expectReport=0;
  return 0x01;
}

int main(void) {
  uchar   updateNeeded = 0;
  uchar   idleCounter = 0;
  uchar i;

/*
  DDRB |= 0x01; // PB0

    while(1)
    {
	PORTB = 0x01;
        _delay_ms(500);
	PORTB = 0x00;
        _delay_ms(500);
    }
*/
  wdt_enable(WDTO_2S); /* Enable watchdog timer 2s */
  hardwareInit(); /* Initialize hardware (I/O) */
  
  odDebugInit();

  usbInit(); /* Initialize USB stack processing */
                                                                                                                                                                                             
  usbDeviceDisconnect(); // enforce re-enumeration                                                                                                                                         
  for(i = 0; i<250; i++) { // wait 500 ms                                                                                                                                                  
     wdt_reset(); // keep the watchdog happy                                                                                                                                              
     _delay_ms(2);                                                                                                                                                                        
  }                                                                                                                                                                                        
  usbDeviceConnect();                                                                                                                                                                      
                                                                                                                                                                                             
  sei(); /* Enable global interrupts */
  
  for(;;){  /* Main loop */
    wdt_reset(); /* Reset the watchdog */
    usbPoll(); /* Poll the USB stack */


    updateNeeded|=scankeys(); /* Scan the keyboard for changes */
    
    /* Check timer if we need periodic reports */
    if(TIFR0 & (1<<TOV0)){
      TIFR0 = 1<<TOV0; /* Reset flag */
      if(idleRate != 0){ /* Do we need periodic reports? */
        if(idleCounter > 3){ /* Yes, but not yet . For 12MHz crystal cahnge to 4 */
          idleCounter -= 4;   /* 16 ms in units of 4 ms. For 12MHz crstal change to 5 */
        }else{ /* Yes, it is time now */
          updateNeeded = 1;
          idleCounter = idleRate;
        }
      }
    }
    
    /* If an update is needed, send the report */
    if(updateNeeded && usbInterruptIsReady()){
      updateNeeded = 0;
      usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
    }
  }
  return 0;
}
